/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <bfd.h>
#include <linux/dprobes.h>
#include <linux/dprobes_in.h>
#include "main.h"

static bfd *abfd;
static asymbol **symtab=(asymbol **) NULL;
static unsigned int symcount;

int symbols_present = 0;
char *kmapfile;

int init_syms(const char * filename)
{
	int retval;
	unsigned int size;
	
	if (symbols_present)
		return 1;

	if (dp_mod_type == DP_MODTYPE_KERNEL) {
		kmapfile = (char *)strdup(filename);
		symbols_present = 1;
		return 0;
	}

	bfd_init();
	bfd_set_default_target("elf32-i386");

	abfd = bfd_openr(filename, NULL);
	if (!abfd) {
		//printf("bfd opening %s failed: symbols are not available.\n");
		goto err;
	}

	retval = bfd_check_format(abfd, bfd_object);
	if (!retval) {
		//printf("file is not of type bfd_object\n");
		goto err;
	}

	size = bfd_get_symtab_upper_bound(abfd);
	if (!size) {
		//printf("symbols are not available.n");
		goto err;
	}

	symtab = (asymbol **)malloc(size);
	if (!symtab) {
		//printf("no memory\n");
		goto err;
	}

	symcount = bfd_canonicalize_symtab(abfd, symtab);
	if (!symcount) {
		goto err;
	}
	symbols_present = 1;
	return 0;
err:
	symbols_present = 0;
	return 1;
}

int cleanup_syms()
{
	if (!symbols_present)
		return 0;

	if (dp_mod_type == DP_MODTYPE_KERNEL) {
		free(kmapfile);
		kmapfile = NULL;
	} else {
		free(symtab);
		if (!bfd_close(abfd)) {
			printf("bfd closed failed\n");
			exit(1);
		}
	}
	symbols_present = 0;
	return 0;
}


/*
 *
 */
asymbol * lookup_object_symbol(const char *name, unsigned int type)
{
	unsigned int i;
	asymbol * s;
	for(i = 0; i < symcount; i++) {
		s = symtab[i];
		if (s->flags & type && !strcmp(s->name, name)) {
			return s;
		}
	}
	return NULL;
}

int lookup_kernel_address(const char *name, unsigned long *val)
{
	FILE *fp=NULL;
	char buff[80];
	int ret;

	sprintf(buff, "awk \' $3 == \"%s\" { print $1 }\' < %s\n",
	name, kmapfile);

	fp = popen(buff, "r");
	if (!fp)
		return 0;

	if (fgets(buff, sizeof(buff), fp)) {
		*val = strtoul(buff, NULL, 16);
		ret = buff[0]? 1: 0;
	} else {
		ret = 0;
	}
		
	pclose(fp);
	return ret;
}

int lookup_function(const char *name, unsigned long * val)
{
	asymbol * s;
	
	if (dp_mod_type & DP_MODTYPE_KERNEL) {
         	if (lookup_kernel_address(name, val)) {
			*val -= PAGE_OFFSET;
			return 1;
		} else {
			return 0;
		}
	} 

	s = lookup_object_symbol(name, BSF_FUNCTION | BSF_GLOBAL);
	if (!s)
		return 0;
	if (dp_mod_type & DP_MODTYPE_KMOD) {
		*val = s->value + s->section->filepos - abfd->sections->filepos;
	} else {
		*val = s->value + s->section->filepos;
	}
	return 1;
}

unsigned long lookup_symbol_value(const char * name)
{
	unsigned long val;
	int retval;
	char errstr[80];
	if (!symbols_present) {
		rpnerror("symbols not available");
		exit(1);
	}

	retval = lookup_function(name, &val);
	if (!retval) {
		sprintf(errstr, "%s: symbol not found ", name);
		rpnerror(errstr);
		exit(1);
	}
	return val;
}

unsigned short get_kmod_align()
{
	struct sec * sect;

	if (!symbols_present) 
		return 0;

	for (sect = abfd->sections; sect; sect = sect->next) {
		if(sect->flags & SEC_CODE)
			return sect->alignment_power;
	}
	return 0;
}
